/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.engine.source;

import java.io.ByteArrayInputStream;
import java.io.Serializable;
import java.security.Principal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.exchange.xml.XmlParser;

import sync4j.foundation.pdi.converter.*;
import sync4j.foundation.pdi.parser.*;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.AbstractSyncSource;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.tools.Base64;

import sync4j.exchange.items.contact.model.Contact;
import sync4j.exchange.items.contact.manager.ContactManager;
import sync4j.exchange.items.contact.ContactParseException;
import sync4j.exchange.DataAccessException;

/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server contacts items
 *
 * @author Fabio Maggi @ Funambol
 * version $Id: ExchangeContactSyncSource.java,v 1.22 2005/06/25 06:56:27 fabius Exp $
 */
public class ExchangeContactSyncSource
extends ExchangeSyncSource
implements SyncSource, Serializable {

    // ----------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml";

    // ----------------------------------------------------------- Private data

    private ContactManager  cm      = null  ;

    // ----------------------------------------------------------- Constructors

    public ExchangeContactSyncSource() {
    }

    // ----------------------------------------------------------- Public Methods

   /**
    * @see SyncSource
    */
    public SyncItem setSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("setSyncItem - source: " + sourceURI);
        }

        Contact         contact        = null  ;
        Contact         newContact     = null  ;

        String          itemKey        = null  ;
        String          href           = null  ;
        String          content        = null  ;

        String          username       = null  ;
        String          credentials    = null  ;

        boolean         isAddContact   = false ;

        try {

            username    = ((Sync4jPrincipal)principal).getUsername()           ;
            credentials = ((Sync4jPrincipal)principal).getEncodedCredentials() ;

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // search href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey);

            if (href == null) {
                //
                // if id not found, new item is created
                // default href = this.CurrentTimesMillis()
                //
                href = String.valueOf(System.currentTimeMillis()) +
                       EXCHANGE_HREF_EXTENSION;
                isAddContact = true;
            }


            if (log.isLoggable(Level.FINEST)) {
                if (!isAddContact) {
                    log.finest(itemKey + " already exists: updating");
                } else {
                    log.finest(itemKey + " is new: adding");
                }
            }

            contact = getContactFromSyncItem(syncItem);
            contact.setHref(href);

            cm = getContactManager();

            newContact = cm.setContact (contact     ,
                                        username    ,
                                        credentials ,
                                        exchangeFolder);

            if (isAddContact) {
                toChangedItems(newContact.getId(), ITEM_ADD);
            } else {
                toChangedItems(newContact.getId(), ITEM_UPDATE);
            }

            syncItem = getSyncItem(newContact);

            return syncItem;

      } catch (DataAccessException e) {

          if (newContact != null) {
              if (isAddContact) {
                  toChangedItems(newContact.getId(), ITEM_ADD_ERROR);
              } else {
                  toChangedItems(newContact.getId(), ITEM_UPDATE_ERROR);
              }
          }

          throw new SyncSourceException("Error setting the item "
                                        + syncItem
                                        , e
                                        );
        }
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList();
        ArrayList ret              = new ArrayList();

        for (int i = 0, l = syncItems.length; i < l; ++i) {
            try {
                ret.add(setSyncItem(principal, syncItems[i]));
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException("Error setting the following items: "
                                         + syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource " +
                       sourceURI              +
                       " - set syncItems");
        }

        return (SyncItem[])ret.toArray(new SyncItem[] {});

    }

    /**
     * @see SyncSource
     */
    public SyncItem getSyncItemFromId(Principal   principal,
                                      SyncItemKey syncItemKey)
    throws SyncSourceException{

        SyncItem[] syncItems =
            getSyncItemsFromIds(principal, new SyncItemKey[] {syncItemKey});

        if ((syncItems == null) || (syncItems.length == 0)) {
            return null;
        }

        return syncItems[0];

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromIds(Principal     principal,
                                          SyncItemKey[] syncItemKeys)
    throws SyncSourceException {

        ArrayList  syncItems   = null ;
        Contact    contact     = null ;

        String     username    = null ;
        String     credentials = null ;

        String     id          = null ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        syncItems = new ArrayList();

        int l = syncItemKeys.length;

        for (int i = 0; ((syncItemKeys != null) && (i < l)); ++i) {

            id = syncItemKeys[i].getKeyAsString();

            try {
                cm = getContactManager();
                contact = cm.getContactById(username      ,
                                            credentials   ,
                                            id            ,
                                            exchangeFolder);
            } catch (DataAccessException e) {

                Throwable previous = e.getCause();

                if (previous instanceof NotFoundException) {

                    Logger log = Sync4jLogger.getLogger("source");
                    if (log.isLoggable(Level.SEVERE)) {
                        log.severe("Contact not found while reading Exchange database: " + e.getMessage());
                    }
                    Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "readExchangeDatabase", e);
                } else{
                    throw new SyncSourceException("Error reading items", e);
                }
            }


            if (contact != null) {
                contact.setStatus(SyncItemState.NEW);
                syncItems.add(getSyncItem(contact));
            }

        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource " +
                       sourceURI              +
                       " - getting syncItems" );
        }

        return (SyncItem[]) syncItems.toArray(new SyncItem[syncItems.size()]);

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getNewSyncItemKeys(Principal principal,
                                            Timestamp since      )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getNewSyncItems(Principal principal,
                                      Timestamp since)
    throws SyncSourceException {

        String ids[] = null;

        ids = this.getNewItemIds();

        if (log.isLoggable(Level.FINEST)) {
            log.severe("Exchange SyncSource "     +
                       sourceURI                  +
                       " - getting new syncItems" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        return filterSyncItems(principal, ids, SyncItemState.NEW);

    }

    /*
     * @see SyncSource
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Principal principal,
                                                Timestamp since      )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Timestamp since      )
    throws SyncSourceException {

        String  [] ids       = null ;
        SyncItem[] syncItems = null ;

        ids = this.getDeleteItemIds();

        int l = ids.length;

        syncItems = new SyncItem[l];

        for (int i = 0; i < l; i++) {

            syncItems[i] = new SyncItemImpl(this, ids[i], SyncItemState.DELETED);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.severe("Exchange SyncSource "         +
                       sourceURI                      +
                       " - getting deleted syncItems" );
        }

        return syncItems;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getUpdatedSyncItems(Principal principal,
                                          Timestamp since)
    throws SyncSourceException {

        SyncItem[] items = null;

        String[] ids = null ;

        ids = this.getUpdateItemIds();

        if (log.isLoggable(Level.FINEST)) {
            log.severe("Exchange SyncSource "         +
                       sourceURI                      +
                       " - getting updated syncItems" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        items = filterSyncItems(principal, ids, SyncItemState.UPDATED);

        return items;
    }


    public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,
                                                Timestamp since)
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public void removeSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        Contact contact      = null  ;

        String  username     = null  ;
        String  credentials  = null  ;

        String  id           = null  ;
        String  href         = null  ;

        id         = syncItem.getKey().getKeyAsString();

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        href = getHref(id);

        try {
            if (href != null) {
                contact = new Contact(id)                          ;
                contact.setHref(href)                              ;
                this.cm.removeContact(contact     ,
                                      username    ,
                                      credentials ,
                                      exchangeFolder)              ;
                toChangedItems(contact.getId(), ITEM_REMOVE)       ;
            } else {
                toChangedItems(contact.getId(), ITEM_REMOVE_ERROR) ;
            }
        } catch (DataAccessException e) {
            toChangedItems(contact.getId(), ITEM_REMOVE_ERROR)     ;
           throw new SyncSourceException("Error reading items", e);
        }

    }

    /**
     * @see SyncSource
     */
    public void removeSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList();

        for (int i=0, l = syncItems.length; i < l; ++i) {
            try {
                removeSyncItem(principal, syncItems[i]);
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException("Error deleting the following items: "
                                         + syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "  +
                       sourceURI               +
                       " - removing syncItems" );
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getAllSyncItems(Principal principal)
    throws SyncSourceException {

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "     +
                       sourceURI                  +
                       " - getting all syncItems" );
        }

        return filterSyncItems(principal, SyncItemState.NEW);
    }

    /**
     * @see SyncSource
     */
    public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItem)
            throws SyncSourceException {

        String username    = ((Sync4jPrincipal) principal).getUsername()           ;
        String credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;
        SyncItem syncTwin      = null;
        try {
            this.cm = getContactManager();
            Contact contact = getContactFromSyncItem(syncItem);
            Contact twin = cm.getContactTwin(contact        ,
                                             username       ,
                                             credentials    ,
                                             exchangeFolder );

            if (twin != null) {
                syncTwin = getSyncItem(twin);
            }

        } catch (DataAccessException ex) {
            throw new SyncSourceException("Error getting twin", ex);
        }

        return syncTwin;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromTwins(Principal principal,
                                            SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList items = new ArrayList();

        for (int i = 0, l = syncItems.length; i < l; ++i){
            items.add(getSyncItemFromTwin(principal, syncItems[i]));
        }

        return (SyncItem[])items.toArray(new SyncItem[items.size()]);

    }

    // --------------------------------------------------------- Private methods


    private Contact getContact(String id      ,
                               String content ,
                               Timestamp t    )
    throws SyncSourceException {

        Contact c = new Contact(id);

        c.setLastUpdate(t) ;

        if (content != null && content.length() > 0) {

            if (this.isEncode()) {
                content = new String (Base64.decode(content));
            }

            if (TYPE_VCARD.equals(getType())) {
                vcard2Contact(content, c);
            } else {
                sifc2Contact(content, c);
            }
        }

        return c;

    }


    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, ids, state
     *
     * @param principal principal
     * @param ids     an array containing contact id
     * @param state   contact state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       String[]   ids       ,
                                       char       state)
    throws SyncSourceException {

       Contact[]  contacts     = null ;

       String     username     = null ;
       String     credentials  = null ;

       username    = ((Sync4jPrincipal) principal).getUsername()           ;
       credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

       try {

           cm = getContactManager();

           contacts = cm.getContacts    (username        ,
                                         credentials     ,
                                         ids             ,
                                         exchangeFolder) ;
        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading contacts: " +
                                          e.getMessage(), e);
        }

        return getSyncItems(contacts, state);

    }


    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, ids, state
     *
     * @param principal principal
     * @param state contact state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       char       state)
    throws SyncSourceException {

        Contact[]  contacts     = null ;

        String     username     = null ;
        String     credentials  = null ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        try {

           cm = getContactManager();

           contacts = cm.getAllContacts (username        ,
                                         credentials     ,
                                         exchangeFolder  ) ;
        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading contacts: " +
                                          e.getMessage(), e);
        }

        return getSyncItems(contacts, state);

    }

    /**
     * Create SyncItem array from Contact array
     *
     * @param contacts
     * @param state
     * @return SyncItem array
     * @trows SyncSourceException
     */
    private SyncItem[] getSyncItems(Contact[] contacts,
                                    char      state)
    throws SyncSourceException {

        SyncItem[] syncItems = null;

        int l = contacts.length;

        syncItems = new SyncItem [l];

        for (int i=0; i < l; i++) {
            contacts [i].setStatus(state)            ;
            syncItems[i] = getSyncItem(contacts[i]) ;
        }

        return syncItems;

    }

    /**
     * Create SyncItem from Contact
     *
     * @param contact
     * @return SyncItem
     * @trows SyncSourceException
     */
    private SyncItem getSyncItem(Contact contact)
    throws SyncSourceException {

        SyncItem syncItem = null;
        String   content  = null;

        syncItem = new SyncItemImpl(this                ,
                                    contact.getId()     ,
                                    contact.getStatus() );

        if (TYPE_VCARD.equals(getType())) {
            content = contact2vcard(contact);
        } else {
            content = contact2sifc(contact);
        }

        if (isEncode()) {
            syncItem.setProperty(
                new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                 Base64.encode((content).getBytes())
            ));

           syncItem.setProperty(
               new SyncProperty(SyncItem.PROPERTY_FORMAT,"b64")
           );

        } else {
            syncItem.setProperty(
                new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                 (content).getBytes())
            );
        }

        return syncItem;

    }

    /**
     * Converts the given contact into a vcard String
     *
     * @param c the contact to convert
     *
     * @return the vcard
     *
     * @throws SyncSourceException in case of convertion errors
     */
    private String contact2vcard(Contact c)
    throws SyncSourceException {
        try {
            return new ContactToVcard(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException( "Conversion error for item "
                                          + c.getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        }
    }

    /**
     * Converts the given contact into a sifc String
     *
     * @param c the contact to convert
     *
     * @return the sifc document
     *
     * @throws SyncSourceException in case of convertion errors
     */
    private String contact2sifc(Contact c)
    throws SyncSourceException {
        try {
            return new ContactToXML(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException( "Convertion error for item "
                                          + c.getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        }
    }

    /**
     *
     * Fill the given Contact object with the properties in the given vcard
     *
     * @param content message in VCard
     * @param c the contect to feed
     *
     * @throws SyncSourceException
     */
    private void vcard2Contact (String content, Contact c)
    throws SyncSourceException {

        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length()==0)) {
            return;
        }

        ByteArrayInputStream  is     = null ;
        VcardParser           parser = null ;

        sync4j.foundation.pdi.contact.Contact contact = null ;

        try {

            /**
             * This is used to replace the String "=\r\n" with the
             * String "\r\n ": This replace is necessary because some devices do
             * not send the correct line delimiter for long lines.
             */
            content = content.replaceAll("=\r\n","\r\n ");

            is = new ByteArrayInputStream (content.getBytes());

            //converting the vcard into a Contact object

            parser = new VcardParser(is, deviceTimeZoneDescr, deviceCharset);

            parser.setLogger(log);

            contact = (sync4j.foundation.pdi.contact.Contact)parser.vCard();

            //
            // We now have to convert a portal contact into a foundation contact
            //
            c.setUid(contact.getUid());
            c.setTimezone(contact.getTimezone());
            c.setNotes(contact.getNotes());
            c.setRevision(contact.getRevision());
            c.setName(contact.getName());
            c.setBusinessDetail(contact.getBusinessDetail());
            c.setPersonalDetail(contact.getPersonalDetail());
            c.setLanguages(contact.getLanguages());
            c.setCategories(contact.getCategories());
            c.setFolder(contact.getFolder());
            c.setSubject(contact.getSubject());
            c.setMileage(contact.getMileage());

            c.setSensitivity(contact.getSensitivity());
            c.setImportance(contact.getImportance());

        } catch (Exception e) {
            e.printStackTrace();
            throw new SyncSourceException( "Convertion error for item "
                                          + c.getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        } finally {
            if (is != null) try { is.close(); } catch (Exception e) {}
        }
    }

    /**
     *
     * Fill the given Contact object with the properties in the given sifc document
     *
     * @param content message in VCard
     * @param c the contect to feed
     *
     * @throws SyncSourceException
     */
    private void sifc2Contact (String content, Contact c)
    throws SyncSourceException {
        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length()==0)) {
            return;
        }


        ByteArrayInputStream  is         = null ;
        XMLContactParser      parser     = null ;

        sync4j.foundation.pdi.contact.Contact contact = null ;

        try {
            is = new ByteArrayInputStream (content.getBytes());

            //converting the vcard into a Contact object

            parser = new XMLContactParser(is);

            contact = (sync4j.foundation.pdi.contact.Contact)parser.parse();

            //
            // We now have to convert a portal contact into a foundation contact
            //
            c.setUid(contact.getUid());
            c.setTimezone(contact.getTimezone());
            c.setNotes(contact.getNotes());
            c.setRevision(contact.getRevision());
            c.setName(contact.getName());
            c.setBusinessDetail(contact.getBusinessDetail());
            c.setPersonalDetail(contact.getPersonalDetail());
            c.setLanguages(contact.getLanguages());
            c.setCategories(contact.getCategories());
            c.setFolder(contact.getFolder());
            c.setSubject(contact.getSubject());
            c.setMileage(contact.getMileage());
            c.setSensitivity(contact.getSensitivity());
            c.setImportance(contact.getImportance());

        } catch (Exception e) {
            throw new SyncSourceException( "Parsing error for item "
                                          + c.getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        } finally {
            if (is != null) try { is.close(); } catch (Exception e) {}
        }
    }

    /**
     *
     * Create ContactManager istance
     * (if not already exist)
     *
     * @trows DataAccessException
     */
    private ContactManager getContactManager()
    throws DataAccessException {

        if (this.cm == null) {
            this.cm = new ContactManager(this.getHost() ,
                                         this.getPort());
        }

        return this.cm;

    }

    /**
     * Creates a new Contact from the given SyncItem
     * @param syncItem SyncItem
     * @return Contact
     * @throws SyncSourceException
     */
    private Contact getContactFromSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Contact contact = null;
        String itemKey = syncItem.getKey().getKeyAsString();

        byte[] itemContent =
            (byte[])syncItem.getPropertyValue(
                SyncItem.PROPERTY_BINARY_CONTENT);

        if (itemContent == null) {
            itemContent = new byte[0];
        }

        String content = new String (itemContent);

        Timestamp t =
             (Timestamp) syncItem.getPropertyValue(SyncItem.PROPERTY_TIMESTAMP);

        contact = getContact(itemKey, content, t);

        return contact;
    }

}
